@make(ais)
@aiInfo(topic=<Indivisible semantic effects>,
number=[AI-00447/02],
classcode=[BI],
statuscode=[WI],
mainref=[09.11@\(11)],
date=[86-12-15])
@aiHeading()
@begin(history)
@begin(toplines)
@mainref()
!class binding interpretation 86-07-10 (provisional classification)
!status work-item 86-08-12
@end(toplines)
!status received 86-07-10
@topic{!topic Indivisible semantic effects}
@end(history)

@subheading[!summary 86-12-15 (DRAFT)]


@subheading[!question 86-08-12 (DRAFT)]

The only case where the Standard explicitly requires an operation to be
implemented as an indivisible operation is found in 9.11(11) with respect to
the direct reading and updating of shared objects.  Does this mean that this
is the only activity that must be implemented in an indivisible manner, or
was it the intent that certain other actions required by the Standard be
implemented in an indivisible manner?  For example, if a task is called
simultaneously from two tasks, presumably the entry call actions are to be
performed in an indivisible manner; in particular, updating the queue of an
entry must presumably be done indivisibly.

What is the required effect if a task is aborted while performing one of the
following actions or if the action is suspended for a period of time and then
resumed?
@begin(quotation)
execution of an abort statement
@begin(quotation1)
if the task executing the abort statement is itself aborted,
might only some of the dependent tasks become abnormal?
@end(quotation1)
activation of tasks
@begin(quotation1)
if the task requiring the activations is aborted,
might only some of the tasks become activated?
@end(quotation1)
creation of an object in a shared collection during evaluation of an
allocator
@begin(quotation1)
suppose two tasks attempt to allocate objects in the same
@begin(quotation1)
collection?
@end(quotation1)
suppose a task is aborted while allocating an object in a shared
@begin(quotation1)
collection?
@end(quotation1)
@end(quotation1)
execution of a regular, conditional or timed entry call
@begin(quotation1)
the actions of updating entry queues
@end(quotation1)
evaluation of the TERMINATED and CALLABLE attributes

completion and termination of a task
@begin(quotation1)
effect on release of storage associated with the task
@end(quotation1)
@end(quotation)
Which of these actions, if any, is required to be implemented in an
indivisible manner?

@subheading[!recommendation 86-07-10 (DRAFT)]


!wording 86-07-10 (DRAFT)


@subheading[!discussion 86-07-10 (DRAFT)]


@begin(appendix)
@subheading[!appendix 86-12-15]
@begin(sectionGroup)
@mybar()

!section 09.11    (11) CRAI-AdaFD       86-06-07                     83-00765
!version 1983
!topic Indivisible operations
@end(sectionGroup)

The only case in which the RM explicitly requires an operation to be
implemented as an indivisible operation is found in 9.11 (11) about
the direct reading and direct updating for objects to which is
applied the pragma SHARED. Does this mean that this is the only
case of operation (or more in general fragment of task activity)
which is required to be implemented as an indivisible activity, or
some other "implicit" case was supposed too obvious to be explicitly
mentioned?

Some possible circumstances under which a fragment of the activity
of a task might be interrupted are the following:
    1) abnormal completion of the task
    2) raising of a STORAGE-ERROR exception
    3) raising of a NUMERIC-ERROR exception
Moreover the activity can be suspended for some time due to:
    4) internal scheduling of the program activity.

In the following we will call a non-safe activity, an activity which,
if interrupted in one of the cases 1) .. 4) might not produce its
"usual effects" (those one mentioned in the RM).
For example normal reading or updating of a variable cannot in
general be considered as a safe activity because from the RM we can
deduce that, if interrupted because of 1) and 4), these operations
might produce some rather partial or inconsistent effects.
Cases 2) and 3) have been included in the above list for sake of
completeness, being not well defined the conditions and the
restrictions under which these exceptions could be raised (but this
is a separate issue, see for example AI-00133/01, AI-00036/02,
and AI-00325/03 AI-00159/00, AI-00325/03 and related).

Let us consider in particular the following fragments of task
activity; which of them can be considered as a safe (probably
implemented as an indivisible) activity?

  a) the execution of an abort statement ?
    (can for example in the cases 1) or 4), or if executed
    concurrently with some new task creations, only some of the
   dependent tasks become abnormal or completed?)

  b) the process of activating some locally created tasks (i.e the
    activity following the successful check for unelaborated bodies
    but preceding the start of the activation of all the required
    tasks)?
    (can for example in the cases 1) or 2) only some of the required
    tasks start their activation?; can the program execution become
    erroneous if some system structure is partially updated?)

  c) the creation of an object in a (shared) collection, during the
   evaluation of an allocator?
   (can for example in the cases 1) or 2) further object creations in
   the same collection result in an erroneous program execution if
   some shared information is partially updated?; can two concurrent
   object allocations produce unpredictable results if the update of
   the system structures recording the collection status is not
   implemented as an indivisible activity?)

  d) the execution of an entry call, or more in general any operation
   on some entry queue?
   (can for example in the cases 1), 2) or 4) further entry calls result
   in an erroneous program execution if some shared information
   (about the queues) is partially updated?; can two concurrent entry
   calls produce unpredictable results if the update of the system
   structures recording the queues status is not implemented as an
   indivisible activity ?)

  e) the evaluation of the 'TERMINATED or 'CALLABLE attribute of a
   task?  the completion or termination of a task?
   (Can the evaluation of such an attribute start an erroneous
   execution or have unpredictable effects, or at least return some
   information which cannot be interpreted as a value of type boolean
   if the evaluation is performed while the task is concurrently
   changing its status?)

We believe that most of these activities should be considered as
safe activities, hence implemented as indivisible activities even in
concurrent or distributed systems;  as the situation seems however
rather confuse some clarification seems required.

@begin(sectionGroup)
@mybar()

!section 09.11    (11) CRAI-AdaFD        86-11-07                    83-00862
!version 1983
!topic Indivisible semantic effects
@end(sectionGroup)

--

The following are some remarks on the current summary of
AI-00447/01.

--

Saying that the effect of aborting a task while executing an abort
statement is "implementation dependent" is probably too much vague.
If the intention is to allow in some cases a "partially executed"
abort statement to have "partial effects" (e.g. only some of the
required task might become abnormal and possibly completed) than
this should be said more explicitly.
If the intention is that the effect of aborting a task while
executing an abort statement might be unpredictable (in the sense
that some system structure might be damaged), than it should be
clearly specified that this possibility is allowed for an
implementation; in this case however the program execution should
not be considered "implementation dependent" but ERRONEOUS.

--

Saying that the effect of aborting a task while creating an object
in a shared collection is implementation dependent is even more
confusing; in this case indeed "partial effects" are difficult to
be imagined. Probably or the object is created correcty, or it is
not created (in both cases the action can be considered as
indivisible), or if the collection structures are partially updated
the program execution should be considered unpredictable (hence
ERRONEOUS, not "implementation dependent").

--

Finally while the RM is clearly vague in making explicit which
actions are to be considered atomic (e.g. it seem rather obvious
that issuing a call should necessarily be consider an indivisible
action, even if not explicitly mentioned in the RM), the RM is
(until now) very precise (at least it tries) in defining which
situation might lead to the beginning of an erroneous execution.
In conclusion we believe that in case of ambiguity it is more
conform to the standard the interpretation which supposes some
activities as indivisible, than the interpretation which add new
cases of erroneous exeutions.

--

Our current attitude in the Ada-FD indeed is to consider both the
creation of an object in a shared collection and the execution of
an abort statement as "safe" activities, allowing in the case of
the abort statement the possibility of having even partial results (when
aborted).

@begin(sectionGroup)
@mybar()

!section 09.10    (05) CRAI-AdaFD        86-11-07                    83-00861
!version 1983
!topic Abort of a task
@end(sectionGroup)

This issue seems partly related with AI-00360 as it deals with the
possible orderings in which the effects of the execution of an
abort statement might occur. It is also partly related to AI-00447
with respect to the definition of some possible "partial effects"
of an abort statement.
If an abort causes a task to become abnormal, completed, and
removes it from an entry queue (see 9.10(5)), can these changes of
status occur at different times, e.g. first becoming abnormal, then
completed, and then removed from the queue?
If Yes (notice that it is easily observable if an abnormal task is
still in an entry queue), can in some cases a rendezvous occur with
an abnormal, but still queued task (or should the called task not
consider the queued but abnormal tasks, or wait until the first
task of the queue, if abnormal, becomes completed and removed from
the queue)?
If task completion might not coincide with task termination (see
AI-00441/01), then 9.10(5) ".. any abnormal task that has not yet
started its activation becomes completed (HENCE ALSO TERMINATED)."
should be interpreted as " any abnormal task that has not yet
started its activation becomes completed AND TERMINATED"?
In this case the task completion and task termination implied by
the abort statement might still occur at different times, is it right?

@begin(sectionGroup)
@mybar()

!section 9.11 (11)  J. Goodenough 86-12-15
!version 1983
!topic LMP/LMC action on AI-00447
@end(sectionGroup)

At its November 1986 meeting, the LMP/LMC decided that this AI needed to be
resolved in conjunction with the general issue of shared variables, and that
no further action will be taken in the immediate future.

@end(appendix)
